#pragma once

#include <string>
#include <functional>
#include <memory>

 // this action will no throw exception in case of failure
#define NOTHROW(action) { try{ action; } catch(...){} }

// generic exception class
class exception_base
{
public:
    exception_base(void) {}
	virtual ~exception_base(void) {}

	virtual std::string to_string(void) const = 0;
};

// send message to event monitor
void hook_exception(const exception_base& e, const std::string& type, const char* filename, unsigned int line);

#define throw_exception(exception_type, ...)       { exception_type e{__VA_ARGS__}; hook_exception(e, typeid(exception_type).name(), __FILE__, __LINE__); throw e; }

// extended debug when triggering exceptions
void ex_debug(const std::string& type, std::function<void(void)> callback);

template<typename type> void ex_debug(std::function<void(void)> callback)
{
	ex_debug(typeid(type).name(), callback);
}

void breakpoint(void);

#define EX_BREAKPOINT(type)	ex_debug<type>(&breakpoint);

/* unknown_exception exception class */
class unknown_exception : public exception_base
{
public:
	virtual std::string to_string(void) const override
	{
		return "Unknown exception!";
	}
};

/* null_pointer_exception exception class */
class null_pointer_exception : public exception_base
{
public:
	virtual std::string to_string(void) const override
	{
		return "Pointer is null!";
	}
};
